class Error(Exception):
    """Base class for pOCCI exceptions."""
    pass


class ParseError(Error):
    """Parse exception."""

    def __init__(self, value, body=None):
        self.value = value
        self.body = body

    def __str__(self):
        if self.body is not None:
            return str(self.value) + ' (%s)' % str(self.body)
        else:
            return str(self.value)


class RenderError(Error):
    """Render exception."""


class TransportError(Error):
    """Transport exception."""


class Generic(dict):
    """ Generic OCCI object in python. """

    """ Memory optimization
    """
    __slots__ = []

    __getattr__ = dict.__getitem__
    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__

    def validate():
        """ Validate OCCI object.

        :return: OCCI object is valid
        :rtype: bool
        """
        return True


class Attribute(Generic):
    """OCCI Attribute Instance

    :ivar string name: attribute name
    :ivar value: attribute value (any type)
    :ivar string type: 'string', 'number', 'boolean', 'enum' [string]
    """
    __slots__ = []


    def validate(self):
        return 'name' in self and 'value' in self


    @classmethod
    def equals(self, a1, a2):
        """Compare two OCCI Attribute Values

        :param occi.Attribute self
        :param {} a1: attribute value
        :param {} a2: attribute value
        :return: attributes equals
        :rtype: bool
        """
        if a1 is None or a2 is None:
            return False
        return 'value' in a1 and 'value' in a2 and a1['value'] == a2['value']


class AttributeDefinition(Generic):
    """OCCI Attribute Definition

    :ivar bool immutable: attribute can't be changed [false]
    :ivar bool required: attribute is required [false]
    :ivar string type: 'string', 'number', 'boolean', 'enum' [string]
    :ivar default: default value
    :ivar string description: description
    """
    __slots__ = []


    def isrequired(self):
        return 'required' in self and self['required']


    def isimmutable(self):
        return 'immutable' in self and self['immutable']


class Category(Generic):
    """OCCI Category

    :ivar string term: term
    :ivar string scheme: scheme
    :ivar string class: class
    :ivar ...: other fields (location, actions, attributes, ...)
    """
    __slots__ = []


    def validate(self):
        return 'term' in self and 'scheme' in self and 'class' in self


class Link(Category):
    """OCCI Link

    :ivar string uri: URI
    :ivar rel string[]: resource types
    :ivar self string: self URI
    :ivar string category[]: types
    :ivar string attributes{}: attributes
    """

    def validate(self):
        return 'uri' in self and 'rel' in self



import json
import os
import sys
import ConfigParser
from collections import OrderedDict

import render


# keep in sync with pOCCI/pOCCI.cfg
occi_defaults = {
    'authtype': 'basic',
    'ignoressl': False,
    'mimetype': 'text/plain',
    'outputformat': 'json',
    'curlverbose': False,
    'connectiontimeout': 60,
    'timeout': 120,

    'occi.tests.category': {'term': 'compute', 'class': 'kind', 'scheme': 'http://schemas.ogf.org/occi/infrastructure#'},
    'occi.tests.entity': {},
}

occi_config = {}
renderers = {}
renderer = None
renderer_big = None
renderer_httpheaders = None


def occi_format(results):
    count_f = 0
    count_o = 0
    for r in results:
        if 'status' in r:
            if r['status']:
                count_o += 1
            else:
                count_f += 1

            r['status'] = result2str(r['status'])

        if 'running_time' in r:
            r['running_time'] = round(r['running_time'], 3)

    out = OrderedDict()
    out['tests'] = results
    out['passed'] = count_o
    out['failed'] = count_f
    return out


def html_escape(s):
    s = str(s)
    s = s.replace("&", "&amp;")
    s = s.replace("<", "&lt;")
    s = s.replace(">", "&gt;")
    s = s.replace("\n", "<br>")
    return s


def occi_print(results, outputformat):
    if outputformat == 'plain':
        for r in results['tests']:
            print '%s  %s' % (r['name'], r['status'])
            if 'reason' in r:
                print >> sys.stderr, r['reason']
    elif outputformat == 'json':
        print json.dumps(results, indent=4)
    elif outputformat in ['html', 'htmltable']:
        if outputformat == 'html':
            print '<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01//EN" "http://www.w3.org/TR/html4/strict.dtd">\n\
<html>\n\
\n\
<head>\n\
    <title>OCCI Compliance Tests Results</title>\n\
    <meta http-equiv="Content-Type" content="text/html;charset=utf-8">\n\
    <style type="text/css">\n\
        th {text-align:left}\n\
        td.ok {color:green}\n\
        td.fail {color:red}\n\
        td.skipped {color:orange}\n\
    </style>\n\
</head>\n\
\n\
<body>\n\
\n\
<table>\n\
    <tr>\n\
        <th>Test</th>\n\
        <th>Running Time</th>\n\
        <th>Status</th>\n\
        <th>Reason</th>\n\
    </tr>'
        for r in results['tests']:
            css = 'skipped'
            if r['status'] == 'OK':
                css = 'ok'
            elif r['status'] == 'FAIL':
                css = 'fail'
            objective = ''
            if 'objective' in r:
                objective = html_escape(r['objective'])
            reason = ''
            if 'reason' in r:
                reason = '\n'.join(r['reason'])
            print '    <tr class="%s">' % css
            print '        <td class="name">%s</td>' % html_escape(r['name'])
            print '        <td class="objective">%s</td>' % html_escape(objective)
            print '        <td class="time">%s</td>' % html_escape(r['running_time'])
            print '        <td class="%s">%s</td>' % (css, html_escape(r['status']))
            print '        <td class="reason">%s</td>' % html_escape(reason)
            print '    </tr>'
        if outputformat == 'html':
            print '</table>\n\
\n\
</body>\n\
\n\
</html>'
    else:
        print >> sys.stderr, 'Only "plain", "json" output types are possible'


def occi_test(name, objective, status, err_msg, running_time=None):
    test = OrderedDict()
    test['name'] = name
    if objective is not None:
        test['objective'] = objective
    test['status'] = status
    if running_time is not None:
        test['running_time'] = running_time
    if err_msg:
        test['reason'] = err_msg

    return test


def result2str(result):
    return 'OK' if result else 'FAIL'


def occi_init():
    """Initialize pOCCI.
    """

    # bigger data requires anything except HTTP Headers renderer
    if occi_config['mimetype'] == 'text/occi':
        occi_config['mimetype.big'] = 'text/plain'
    else:
        occi_config['mimetype.big'] = occi_config['mimetype']

    occi_render_init()


def occi_config_init():
    """Initialize pOCCI configuration.

    Reads the configuration file: /etc/pOCCI.cfg, ~/.pOCCI.cfg.
    """
    global occi_config

    config = ConfigParser.ConfigParser()
    config.read(['/etc/pOCCI.cfg', os.path.expanduser('~/.pOCCI.cfg')])
    if config.has_section('main'):
        for key, value in config.items('main'):
            #print 'config: %s = %s (%s)' % (key, value, type(eval(value)))
            occi_config[key] = eval(value)

    for key, value in occi_defaults.iteritems():
        if key not in occi_config:
            occi_config[key] = value

    return True


def occi_render_init():
    """Initialize pOCCI renderers.

    Limitations:
       - For HTTP GET requests 'text/occi' is always needed
       - For bigger data 'text/occi' should not be used (using 'text/plain')
    """
    mimetypes = ['text/plain', 'text/occi']
    self = sys.modules[__name__]

    # renderers always needed
    for mime in mimetypes:
        renderers[mime] = render.create_renderer(mime)

    # user configurable renderer
    if occi_config['mimetype'] in mimetypes:
        renderer = renderers[occi_config['mimetype']]
    else:
        renderer = render.create_renderer(occi_config['mimetype'])

    # big data requires anything except HTTP Headers renderer
    renderer_big = renderer
    if occi_config['mimetype'] != occi_config['mimetype.big']:
        if occi_config['mimetype.big'] in mimetypes:
            renderer_big = renderers[occi_config['mimetype.big']]
        else:
            renderer_big = render.create_renderer(occi_config['mimetype.big'])

    # HTTP GET requests needs HTTP Headers renderer
    renderer_httpheaders = renderer
    if occi_config['mimetype'] != 'text/occi':
        renderer_httpheaders = renderers['text/occi']

    self.renderer = renderer
    self.renderer_big = renderer_big
    self.renderer_httpheaders = renderer_httpheaders


if not occi_config:
    occi_config_init()



import occi
from render_base import Renderer, check_url


class URIListRenderer(Renderer):
    """URI list OCCI Renderer

    Empty array is always returned as headers during rendering.
    """

    def render_categories(self, categories):
        """Render OCCI Category collection

        This method can't be used in URI list rendering.

        :param occi.Category category[]: OCCI Category array
        :return: render result
        :rtype: [string, string[]]
        """
        raise occi.RenderError('This method can\'t be used with URI list rendering.')


    def render_resource(self, categories, links=None, attributes=None):
        """Render OCCI Resource instance

        This method can't be used in URI list rendering.

        :param occi.Category categories[]: OCCI Category array
        :param occi.Link links[]: OCCI Link array
        :param occi.Attribute attributes[]: OCCI Attribute array
        :return: render result
        :rtype: [string, string[]]
        """
        raise occi.RenderError('This method can\'t be used with URI list rendering.')


    def render_locations(self, locations):
        """ Render Locations

        :param string location[]: location URI
        :return: render result
        :rtype: [string, string[]]
        """
        return ['\n'.join(locations), []]


    def parse_categories(self, body, headers):
        """Parse OCCI Category Collection

        This method can't be used in URI list rendering.

        :param string body[]: text to parse
        :param string headers[]: headers to parse
        :return: Array of OCCI Categories
        :rtype: occi.Category[]
        """
        raise occi.ParseError('This method can\'t be used with URI list rendering.')


    def parse_locations(self, body, headers):
        """Parse OCCI Entity collection

        :param string body[]: text to parse
        :param string headers[]: headers to parse
        :return: array of renderer-specific strings
        :rtype: string[]
        """
        locations = []
        for uri in body:
            if not check_url(uri, scheme=True, host=True):
                raise occi.ParseError('Invalid URI in OCCI Entity collection', uri)
            locations.append(uri)

        return locations


    def parse_resource(self, body, header):
        """Parse OCCI Resource instance

        This method can't be used in URI list rendering.

        :param string body[]: text to parse
        :param string headers[]: headers to parse
        :return: categories, links, and attributes
        :rtype: [occi.Category categories[], occi.Link links[], occi.Attribute attributes[]]
        """
        raise occi.ParseError('This method can\'t be used with URI list rendering.')



import re

import occi

from render_base import Renderer
from render_text import *


class HTTPHeadersRenderer(TextRenderer):
    """HTTP Headers OCCI Renderer

    RFC 7230 http://www.ietf.org/rfc/rfc7230.txt.

    Empty string is always returned as body during rendering.

    Beware of HTTP Headers size limitations. It is better to not use 'text/occi' mimetype for transfering OCCI Category Collection.
    """

    reHeaderChunk = re.compile(r'(?P<chunk>((?P<quoted>"([^"\\]|\\.)*")|(?P<term>[^,"\\]+))*)(?P<sep>\s*,\s*)?')

    def render_category(self, category):
        """Render OCCI Category

        :param occi.Category category: OCCI Category object
        :return: render result
        :rtype: [string, string[]]
        """
        return ['', ['Category: ' + text_category(category)]]


    def render_categories(self, categories):
        """Render OCCI Category collection

        :param occi.Category category[]: OCCI Category array
        :return: render result
        :rtype: [string, string[]]
        """
        if not categories:
            return ['', []]
        res = []
        for category in categories:
            res.append(text_category(category))
        return ['', ['Category: ' + ','.join(res)]]


    def render_links(self, links):
        """ Render OCCI Links

        :param occi.Link link[]: OCCI Link array
        :return: render result
        :rtype: string[]
        """
        if not links:
            return []
        res = []
        for link in links:
            res.append(text_link(link))
        return ['Link: ' + ','.join(res)]


    def render_attributes(self, attributes):
        """Render Attributes

        :param occi.Attribute attribute[]: OCCI Attribute object
        :return: render result
        :rtype: string[]
        """
        if not attributes:
            return []

        s = []
        for attribute in attributes:
            s.append(text_attribute_repr(attribute))
        return ['X-OCCI-Attribute: ' + ','.join(s)]


    def render_locations(self, locations):
        """ Render Locations

        :param string location: location URI
        :return: render result
        :rtype: [string, string[]]
        """
        return ['', ['X-OCCI-Location: ' + ','.join(locations)]]


    def render_resource(self, categories, links=None, attributes=None):
        """Render OCCI Resource instance

        :param occi.Category category: OCCI Category object
        :param occi.Link links[]: OCCI Link array
        :param occi.Attribute attributes[]: OCCI Attribute array
        :return: render result
        :rtype: [string, string[]]
        """
        Renderer.render_resource(self, categories, links, attributes)
        res = []
        cat_s, cat_h = self.render_categories(categories)
        res += cat_h
        if links is not None:
            res += self.render_links(links)
        if attributes is not None:
            res += self.render_attributes(attributes)
        return ['', res]


    def parse_header_values(self, body):
        """Helper generator method to split values in HTTP Header.

        :param string body: parsed text
        :return: result values
        :rtype: string
        """
        m = None
        while body:
            m = HTTPHeadersRenderer.reHeaderChunk.match(body)
            #print m
            #if m:
            #    print m.groupdict()
            #    print m.end()
            # error on empty match
            if m is None or m.group('chunk') is None or m.end() == 0:
                raise occi.ParseError('Bad quoting in HTTP Headers', body)
                m = None
                break
            chunk = m.group('chunk')
            body = body[m.end():]
            if body and not m.group('sep'):
                raise occi.ParseError('Separator expected in HTTP Headers (%s)' % chunk, body)
                m = None
                break
            yield chunk.strip(' \t')
            #print 'remains: #%s#' % body
        if m and m.group('sep'):
            yield ''


    def parse_categories(self, body, headers):
        """Parse OCCI Category Collection

        Beware of HTTP Headers size limitations. It is better to not use 'text/occi' mimetype for transfering OCCI Category Collection.

        :param string body[]: text to parse (unused in plain/occi)
        :param string headers[]: headers to parse
        :return: Array of OCCI Categories
        :rtype: occi.Category[]
        """
        categories = []
        category_ids = set()

        for line in headers:
            line = line.rstrip('\r\n')
            matched = TextRenderer.reCategory.match(line)
            if not matched:
                continue

            line = line[matched.end():]
            #print 'CATEGORY HIT:'
            #print line
            bodies = self.parse_header_values(line)
            #print 'SPLAT BODIES:'
            #print '\n\n'.join(bodies)

            for cat_s in bodies:
                # use the helper parser function inherited from text/plain renderer
                category = TextRenderer.parse_category_body(self, cat_s)

                # check uniqueness
                key = category['term'] + category['scheme']
                if key in category_ids:
                    raise occi.ParseError('Category not unique (term "%s", scheme "%s")' % (category['term'], category['scheme']), cat_s)
                category_ids.add(key)

                categories.append(category)

        return categories


    def parse_locations(self, body, headers):
        """Parse OCCI Entity collection

        :param string body[]: text to parse (unused in text/occi)
        :param string headers[]: headers to parse
        :return: Array of links
        :rtype: string[]
        """
        locations = []
        for line in headers:
            line = line.rstrip('\r\n')
            matched = TextRenderer.reLocation.match(line)
            if not matched:
                continue
            uris_str = matched.group(2)
            uris = self.parse_header_values(uris_str)
            for uri in uris:
                if not check_url(uri, scheme=True, host=True):
                    raise occi.ParseError('Invalid URI in OCCI Entity collection', line)
                locations.append(uri)

        return locations


    def parse_resource(self, body, headers):
        """Parse OCCI Resource instance

        :param string body[]: text to parse (unused in text/occi)
        :param string headers[]: headers to parse
        :return: categories, links, and attributes
        :rtype: [occi.Category categories[], occi.Link links[], occi.Attribute attributes[]]
        """
        categories = []
        links = []
        attributes = []

        for line in headers:
            line = line.rstrip('\r\n')

            matched = TextRenderer.reCategory.match(line)
            if matched is not None:
                cats_str = line[matched.end():]
                for s in self.parse_header_values(cats_str):
                    categories.append(self.parse_category_body(s))
                continue

            matched = TextRenderer.reLink.match(line)
            if matched is not None:
                links_str = line[matched.end():]
                for s in self.parse_header_values(links_str):
                    links.append(self.parse_link_body(s))
                continue

            matched = TextRenderer.reAttribute.match(line)
            if matched is not None:
                attrs_str = line[matched.end():]
                for s in self.parse_header_values(attrs_str):
                    attributes.append(self.parse_attribute_body(s))
                continue
            else:
                continue

        return [categories, links, attributes]



import json
import pycurl
import re
import sys
if sys.version_info >= (3,):
    from StringIO import BytesIO
else:
    from StringIO import StringIO

import occi


# for curl helper callback function
header = []


def get_header2(buff):
    global header
    header.append(buff)


def get_header3(buff):
    global header
    header.append(buff.decode('iso-8859-1'))


class Transport:
    """Transport base class. Curl is used, keystone authentication supported.

    Available methods: delete(), get(), post(), put().
    """

    reEncoding = re.compile(r';\s*charset=(\S+)')
    reHeader = re.compile(r'([^:]*)\s*:\s*(.*)')
    reStatus = re.compile(r'^HTTP')

    def dprint(self, s):
        if self.verbose:
            print '[pOCCI.curl] %s' % s


    def __init__(self, config):
        self.auth = {}
        self.authtype = config['authtype']
        self.config = config
        self.retry = False
        self.verbose = False
        if 'curlverbose' in config:
            self.verbose = config['curlverbose']

        if self.authtype == 'basic':
            if 'user' not in config:
                raise occi.TransportError('User and password is required for "basic" authentication')
        elif self.authtype == 'x509':
            if 'cert' not in config:
                raise occi.TransportError('SSL certificate and key is required for "x509" authentication')

        self.curl = pycurl.Curl()
        curl = self.curl
        curl.setopt(pycurl.VERBOSE, self.verbose)
        if 'connectiontimeout' in config:
            curl.setopt(pycurl.CONNECTTIMEOUT, config['connectiontimeout'])
        if 'timeout' in config:
            curl.setopt(pycurl.TIMEOUT, config['timeout'])

        if 'capath' in config and config['capath']:
            curl.setopt(pycurl.CAPATH, config['capath'])
        if 'cachain' in config and config['cachain']:
            curl.setopt(pycurl.CAINFO, config['cachain'])

        # Disable check of SSL certificate
        if 'ignoressl' in config and config['ignoressl']:
            curl.setopt(pycurl.SSL_VERIFYPEER, 0)
            curl.setopt(pycurl.SSL_VERIFYHOST, 0)

        self.dprint('authtype: %s' % self.authtype)
        if self.authtype == 'basic':
            user = self.config['user']
            password = ''
            if 'passwd' in self.config:
                password = self.config['passwd']
            curl.setopt(pycurl.HTTPAUTH, pycurl.HTTPAUTH_BASIC)
            curl.setopt(pycurl.USERPWD, '%s:%s' % (user, password))
            self.dprint('user: %s' % user)
        elif self.authtype == 'x509':
            if 'cert' in config and config['cert']:
                curl.setopt(pycurl.SSLCERT, config['cert'])
                self.dprint('cert: %s' % config['cert'])
            if 'key' in config and config['key']:
                curl.setopt(pycurl.SSLKEY, config['key'])
                self.dprint('key: %s' % config['key'])
            if 'passphrase' in config and config['passphrase']:
                curl.setopt(pycurl.SSLCERTPASSWD, config['passphrase'])

        # HTTP header response
        if sys.version_info >= (3,):
            curl.setopt(pycurl.HEADERFUNCTION, get_header3)
        else:
            curl.setopt(pycurl.HEADERFUNCTION, get_header2)


    def clean(self):
        self.retry = False

        curl = self.curl
        curl.unsetopt(pycurl.CUSTOMREQUEST)
        curl.setopt(pycurl.HTTPHEADER, [])
        curl.setopt(pycurl.POST, False)


    def perform(self):
        global header

        curl = self.curl

        if sys.version_info >= (3,):
            buffer = BytesIO()
        else:
            buffer = StringIO()
        curl.setopt(pycurl.WRITEFUNCTION, buffer.write)

        header = []
        try:
            curl.perform()
        except pycurl.error as pe:
            raise occi.TransportError(pe)

        ## 'Server: Apache/2.2.22 (Debian)\r\n'
        h = {}
        http_status = None
        for item in header:
            m = Transport.reHeader.match(item.rstrip())
            if m and m.groups >= 2:
                key = m.group(1)
                value = m.group(2)
                h[key.lower()] = value
            else:
                if Transport.reStatus.match(item):
                    http_status = item.rstrip()
        content_type = None
        if 'content-type' in h:
            content_type = re.split(';', h['content-type'])[0]

        body = buffer.getvalue()
        buffer.close()
        if sys.version_info >= (3,):
            encoding = 'iso-8859-1'
            if content_type:
                match = Transport.reEncoding.search(h['content-type'])
                if match:
                    encoding = match.group(1)
            body = body.decode(encoding)

        return [body, header, http_status, content_type, h]


    def auth_keystone(self, url, tenants=True):
        if not url.endswith('/'):
            url += '/'
        self.auth['url'] = url
        self.dprint('Keystone URL: %s' % self.auth['url'])

        version = 'v2.0'
        url += version

        if self.authtype == 'basic':
            user = self.config['user']
            password = ''
            if 'passwd' in self.config:
                password = self.config['passwd']
            body = {
                'auth': {
                    'passwordCredentials': {
                        'username': user,
                        'password': password,
                    },
                },
            }
        elif self.authtype == 'x509':
            body = {
                'auth': {
                    'voms': True,
                },
            }
            if 'keystone' in self.config:
                body['auth']['tenantName'] = self.config['keystone']

        curl = self.curl

        self.clean()
        self.retry = True
        curl.setopt(pycurl.HTTPHEADER, ['Content-Type: application/json'])
        curl.setopt(pycurl.URL, url + '/tokens')
        curl.setopt(pycurl.POST, 1)
        body = json.dumps(body, indent=4)
        curl.setopt(pycurl.POSTFIELDS, body)
        self.dprint('Keystone sending: %s' % body)

        body, header_list, http_status, content_type, header = self.perform()
        if self.verbose:
            self.dprint('Keystone result: %s' % http_status)
            #self.dprint('  headers: ' + str(header))
            #self.dprint('  body: ' + body)
        if re.match(r'200 OK', http_status) is not None:
            raise occi.TransportError('Keystone failed: %s' % http_status)

        keystone = json.loads(body)
        if 'access' not in keystone or 'token' not in keystone['access'] or 'id' not in keystone['access']['token']:
            raise occi.TransportError("Can't get keystone token from: %s" % body)
        self.auth['token'] = keystone['access']['token']['id']

        if tenants and 'tenants' not in self.auth:
            # request tenants, if not already in the response
            if 'tenant' not in keystone['access']['token']:
                self.clean()
                self.retry = True
                curl.setopt(pycurl.HTTPHEADER, ['Content-Type: application/json', 'X-Auth-Token: %s' % self.auth['token']])
                curl.setopt(pycurl.URL, url + '/tenants')

                self.dprint('Keystone exploring tenants')
                body, header_list, http_status, content_type, header = self.perform()
                #self.dprint('  ==> body: %s' % body)
                tenants = json.loads(body)
                #self.dprint('  ==> json: %s' % tenants['tenants'])
                self.auth['tenants'] = tenants['tenants']
            else:
                self.auth['tenants'] = [keystone['access']['token']['tenant']]

            self.auth['tenants_list'] = []
            for t in self.auth['tenants']:
                if 'enabled' not in t or t['enabled']:
                    self.auth['tenants_list'] += [t['name']]
            self.dprint('  ==> tenants: %s' % ','.join(self.auth['tenants_list']))


    def request(self, url=None, mimetype=None, headers=[]):
        if url is None:
            url = self.config['url']
            if not url.endswith('/'):
                url += '/'
            url += '-/'

        if mimetype is None and 'mimetype' in self.config:
            mimetype = self.config['mimetype']

        curl = self.curl

        curl.setopt(pycurl.URL, url)

        # Keystone
        if 'token' in self.auth:
            headers = ['X-Auth-Token: %s' % self.auth['token']] + headers

        # Set appropriate mime type
        if mimetype:
            headers = ['Accept: %s' % mimetype] + headers
        else:
            headers = ['Accept: */*'] + headers

        # Set requested HTTP headers
        if headers:
            curl.setopt(pycurl.HTTPHEADER, headers)

        body, header_list, http_status, content_type, header = self.perform()
        self.dprint('First request status: %s' % http_status)

        if re.match(r'HTTP/.* 401 .*', http_status) and 'www-authenticate' in header:
            self.dprint('WWW-Authenticate extension detected')
            m = re.match(r'Keystone uri=\'([^\']*)\'', header['www-authenticate'])
            if m and m.group(1):
                self.dprint('Keystone detected')
                self.auth_keystone(url=m.group(1))
            return [None, None, http_status, content_type]
        else:
            return [body.splitlines(), header_list, http_status, content_type]


    def delete(self, url=None, mimetype=None, headers=[], body=None):
        """Send HTTP DELETE request

        :param string base_url: OCCI server URL (default: from config)
        :param string url: URL
        :param string mimetype: accepted mimetype (empty string='\*/\*')
        :param string headers[]: HTTP Headers

        :return: [body, header, HTTP status, content type]
        :rtype: [string[], string[], string, string]
        """
        self.clean()

        curl = self.curl
        if body:
            curl.setopt(pycurl.POST, 1)
            curl.setopt(pycurl.POSTFIELDS, body)
            if self.verbose:
                print "[pOCCI.curl] === DELETE ==="
                print body
                print "[pOCCI.curl] =============="
        curl.setopt(pycurl.CUSTOMREQUEST, 'DELETE')
        l = self.request(url=url, mimetype=mimetype, headers=headers)

        if self.retry:
            self.dprint('repeating the DELETE request...')
            self.clean()
            curl.setopt(pycurl.CUSTOMREQUEST, 'DELETE')
            l = self.request(url=url, mimetype=mimetype, headers=headers)

        return l


    def get(self, url=None, mimetype=None, headers=[]):
        """Send HTTP GET request

        :param string base_url: OCCI server URL (default: from config)
        :param string url: URL
        :param string mimetype: accepted mimetype (empty string='\*/\*')
        :param string headers[]: HTTP Headers

        :return: [body, header, HTTP status, content type]
        :rtype: [string[], string[], string, string]
        """
        self.clean()
        l = self.request(url=url, mimetype=mimetype, headers=headers)

        if self.retry:
            self.dprint('repeating the GET request...')
            self.clean()
            l = self.request(url=url, mimetype=mimetype, headers=headers)

        return l


    def post(self, url=None, mimetype=None, headers=[], body='OK'):
        """Send HTTP POST request

        :param string base_url: OCCI server URL (default: from config)
        :param string url: URL
        :param string mimetype: accepted mimetype (empty string='\*/\*')
        :param string headers[]: HTTP Headers
        :param string post: HTTP Body

        :return: [body, header, HTTP status, content type]
        :rtype: [string[], string[], string, string]
        """
        self.clean()

        curl = self.curl
        curl.setopt(pycurl.POST, 1)
        curl.setopt(pycurl.POSTFIELDS, body)
        if self.verbose:
            print "[pOCCI.curl] ==== POST ===="
            print body
            print "[pOCCI.curl] =============="
        l = self.request(url=url, mimetype=mimetype, headers=headers)

        if self.retry:
            self.dprint('repeating the POST request...')
            self.clean()
            curl.setopt(pycurl.POST, 1)
            curl.setopt(pycurl.POSTFIELDS, body)
            l = self.request(url=url, mimetype=mimetype, headers=headers)

        return l


    def put(self, url=None, mimetype=None, headers=[], body='OK'):
        """Send HTTP POST request

        :param string base_url: OCCI server URL (default: from config)
        :param string url: URL
        :param string mimetype: accepted mimetype (empty string='\*/\*')
        :param string headers[]: HTTP Headers
        :param string post: HTTP Body

        :return: [body, header, HTTP status, content type]
        :rtype: [string[], string[], string, string]
        """

        self.clean()

        curl = self.curl
        curl.setopt(pycurl.CUSTOMREQUEST, 'PUT')
        curl.setopt(pycurl.POST, 1)
        curl.setopt(pycurl.POSTFIELDS, body)
        if self.verbose:
            print "[pOCCI.curl] ==== PUT ===="
            print body
            print "[pOCCI.curl] ============="
        l = self.request(url=url, mimetype=mimetype, headers=headers)

        if self.retry:
            self.dprint('repeating the PUT request...')
            self.clean()
            curl.setopt(pycurl.CUSTOMREQUEST, 'PUT')
            curl.setopt(pycurl.POST, 1)
            curl.setopt(pycurl.POSTFIELDS, body)
            l = self.request(url=url, mimetype=mimetype, headers=headers)

        return l



import re
import sys
import time
import urlparse

from occi_libs import *
import transport
import occi


renderer = None
renderer_big = None
renderer_httpheaders = None
connection = None

#'Category: offline;scheme="http://schemas.ogf.org/occi/infrastructure/storagelink/action#";class="action";title="deactivate storagelink"'
required_categories = [
    occi.Category({'scheme': 'http://schemas.ogf.org/occi/core#', 'term': 'entity'}),
    occi.Category({'scheme': 'http://schemas.ogf.org/occi/core#', 'term': 'resource'}),
    occi.Category({'scheme': 'http://schemas.ogf.org/occi/core#', 'term': 'link'}),
]

example_attributes = {
    'occi.core.id': occi.Attribute({'name': 'occi.core.id', 'value': '1'}),
    'occi.core.title': occi.Attribute({'name': 'occi.core.title', 'value': 'Test_title_%d' % time.time()}),
    'occi.storage.size': occi.Attribute({'name': 'occi.storage.size', 'type': 'number', 'value': 0.1}),
    'occi.storage.state': occi.Attribute({'name': 'occi.storage.state', 'type': 'enum', 'value': "inactive"}),
    'occi.storagelink.deviceid': occi.Attribute({'name': 'occi.storagelink.deviceid', 'type': 'string', 'value': "/dev/blk0"})
}


class Test:
    """Base class for OCCI compliance tests"""

    objective = None
    categories = []


    @classmethod
    def test(self=None):
        return [False, ['Not implemented']]


    @classmethod
    def fetch_categories(self, err_msg):
        categories = []
        check_parse = True

        body, response_headers, http_status, content_type = connection.get(mimetype=occi_config['mimetype.big'])

        if body is not None:
            try:
                categories = renderer_big.parse_categories(body, response_headers)
            except occi.ParseError as pe:
                check_parse = False
                err_msg.append(str(pe))

        if occi_config['curlverbose']:
            print '==== CATEGORIES ===='
            if check_parse:
                for category in categories:
                    print category
            else:
                print 'ERROR: %s' % err_msg[-1]
            print '===================='

        # for OpenStack
        # TODO: needed?
        for category in categories:
            if 'location' not in category:
                category['location'] = '/' + category['term'] + '/'

        Test.categories = categories
        return [check_parse, body, response_headers, http_status, content_type]


    @classmethod
    def get_category(self, uri):
        for cat in Test.categories:
            if cat['scheme'] + cat['term'] == uri:
                return cat
        return None


    @classmethod
    def clear_categories(self):
        Test.categories = []


    @classmethod
    def search_category(self, filter):
        for cat in Test.categories:
            if match_category(cat, filter):
                return cat
        return None


    @classmethod
    def pretest_http_status(self, http_ok_status, err_msg, force=False):
        check_pretest = True
        check_categories = True
        body = None
        response_headers = None
        http_status = None
        content_type = None

        if not Test.categories or force:
            check_categories, body, response_headers, http_status, content_type = Test.fetch_categories(err_msg)
            check_pretest, tmp_err_msg = check_http_status(http_ok_status, http_status)
            err_msg += tmp_err_msg
        return [body, response_headers, http_status, content_type, check_pretest and check_categories]


def testsuite_init():
    """Initialize OCCI testsuite

    Renderers from occi_libs needs to be initialized first.
    """
    self = sys.modules[__name__]
    if 'pOCCI.occi_libs' in sys.modules:
        occi_libs = sys.modules['pOCCI.occi_libs']
    else:
        occi_libs = sys.modules['occi_libs']

    self.renderer = occi_libs.renderer
    self.renderer_big = occi_libs.renderer_big
    self.renderer_httpheaders = occi_libs.renderer_httpheaders

    if not self.renderer:
        print >> sys.stderr, 'No renderer (invalid mimetype?)'
        sys.exit(2)

    self.connection = transport.Transport(occi_config)


def check_content_type(content_type):
    if content_type in ['text/occi', 'text/plain', 'application/occi+json']:
        return [True, []]
    else:
        return [False, ['Wrong Content-Type in response']]


def check_requested_content_type(content_type, big=False, headers=False):
    if big:
        requested = occi_config['mimetype.big']
    elif headers:
        requested = 'text/occi'
    else:
        requested = occi_config['mimetype']

    if content_type == requested:
        return [True, []]
    else:
        return [False, ['Result mime type differs']]


def check_http_status(http_expected_status, http_status):
    if not re.match(r'^HTTP/.* %s' % http_expected_status, http_status):
        return [False, ['HTTP status is not %s (%s)' % (http_expected_status, http_status)]]
    else:
        return [True, []]


def match_category(category, filter):
    for key, value in filter.items():
        if not (key in category and category[key] == value):
            return False
    return True


def match_entity(attributes, filter):
    adict = {}
    for a in attributes:
        adict[a['name']] = a
    for key, value in filter.items():
        a2 = occi.Attribute({'value': value})
        if key not in adict or not occi.Attribute.equals(adict[key], a2):
            #print '[match_entity] bad key %s' % key
            return False

    return True


def check_body_entities(body, headers, err_msg=[]):
    entities = []
    try:
        entities = renderer_httpheaders.parse_locations(body, headers)
    except occi.ParseError as pe:
        err_msg.append(str(pe))
        return None

    if not entities:
        try:
            entities = renderer.parse_locations(body, headers)
        except occi.ParseError as pe:
            err_msg.append(str(pe))
            err_msg.append('HTTP Body doesn\'t contain the OCCI Compute Resource description')
            return None

    return entities


def gen_id(prefix):
    return '%s_%d' % (prefix, time.time())


class CORE_DISCOVERY001(Test):
    objective = 'Retrieving all OCCI Categories supported by the OCCI Server'

    @classmethod
    def test(self=None):
        check_cat = False

        err_msg = []

        body, response_headers, http_status, content_type, check_pretest = Test.pretest_http_status("200 OK", err_msg, force=True)

        check_ct, tmp_err_msg = check_content_type(content_type)
        err_msg += tmp_err_msg

        check_rct, tmp_err_msg = check_requested_content_type(content_type, big=True)
        err_msg += tmp_err_msg

        count = 0
        for category in required_categories:
            if Test.search_category(category) is not None:
                count += 1

        if count == len(required_categories):
            check_cat = True
        else:
            err_msg.append('Body doesn\'t contain appropriate categories')

        return [check_pretest and check_ct and check_rct and check_cat, err_msg]


class CORE_DISCOVERY002(Test):
    objective = 'Retrieving the OCCI Categories with an OCCI Category filter from the OCCI Server'

    @classmethod
    def test(self=None):
        err_msg = []
        filtered_categories = []

        check = True

        body, response_headers, http_status, content_type, check_pretest = Test.pretest_http_status("200 OK", err_msg)
        if not Test.categories:
            err_msg += ['No categories returned']
            return [False, err_msg]

        filter = occi.Category({
            'term': Test.categories[0]['term'],
            'scheme': Test.categories[0]['scheme'],
            'class': Test.categories[0]['class'],
        })
        cat_in = []
        cat_in.append('Content-Type: text/occi')
        cat_in += renderer_httpheaders.render_category(filter)[1]

        body, response_headers, http_status, content_type = connection.get(headers=cat_in)

        check_ct, err_msg = check_content_type(content_type)

        if not re.match(r'^HTTP/.* 200 OK', http_status):
            check = False
            err_msg.append('HTTP status on filter not 200 OK (%s)' % http_status)

        check_rct, tmp_err_msg = check_requested_content_type(content_type)
        err_msg += tmp_err_msg

        try:
            filtered_categories = renderer.parse_categories(body, response_headers)
        except occi.ParseError as pe:
            check = False
            err_msg.append(str(pe))

        check_filter = False
        for cat in filtered_categories:
            if match_category(cat, filter):
                check_filter = True
                break
        if not check_filter:
            err_msg.append('Category "%s" (scheme "%s") not in filtered result' % (filter['term'], filter['scheme']))

        return [check and check_pretest and check_ct and check_rct and check_filter, err_msg]


class CORE_CREATE001(Test):
    """Create an OCCI Resource

    Unsupported test: Creating compute instances without os_tpl is not supported.

    It can be called by::
       pOCCI -t 'CORE/CREATE/001'
    """

    objective = 'Create an OCCI Resource'

    @classmethod
    def test(self=None):
        err_msg = []

        body, response_headers, http_status, content_type, check_pretest = Test.pretest_http_status("200 OK", err_msg)

        if not check_pretest:
            return [False, err_msg]

        #kind = Test.search_category({'class': 'kind'})
        kind = Test.search_category({'class': 'kind', 'term': 'compute'})
        #print kind

        if kind:
            for item in ['location', 'term', 'scheme']:
                if item not in kind.keys():
                    err_msg.append('No %s in OCCI Kind' % item)
                    return [False, err_msg]
        else:
            err_msg.append('No OCCI Kind found')
            return [False, err_msg]

        new_cat_s, new_cat_h = renderer.render_resource(
            categories=[
                occi.Category({
                    'term': 'compute',
                    'scheme': 'http://schemas.ogf.org/occi/infrastructure#',
                    'class': 'kind',
                    'title': 'titulek',
                }),
            ],
            attributes=[
                occi.Attribute({
                    'name': 'occi.core.id',
                    'value': gen_id('Compute'),
                }),
                occi.Attribute({
                    'name': 'occi.core.title',
                    'value': 'titulek',
                }),
                occi.Attribute({
                    'name': 'occi.core.summary',
                    'value': 'sumarko',
                }),
                occi.Attribute({
                    'name': 'occi.compute.architecture',
                    'value': 'arch',
                }),
            ]
        )

        body, response_headers, http_status, content_type = connection.post(url=occi_config['url'] + kind['location'], headers=['Content-Type: %s' % occi_config['mimetype']] + new_cat_h, body=new_cat_s)
        Test.clear_categories()
        check_create, tmp_err_msg = check_http_status("201 Created", http_status)
        err_msg += tmp_err_msg

        if not check_create:
            print >> sys.stderr, body

        return [check_create, err_msg]


class CORE_CREATE006(Test):
    """Add an OCCI Mixin definition

    Unsupported test: Not implemented.

    It can be called by::
       pOCCI -t 'CORE/CREATE/006'
    """

    objective = 'Add an OCCI mixin definition'

    @classmethod
    def test(self=None):
        err_msg = []
        new_mixin = 'Category: stufik; scheme="http://example.com/occi/my_stuff#"; class="mixin"; location="/mixin/resource_tpl/extra_large/"; rel="http://schemas.ogf.org/occi/infrastructure#resource_tpl"'
        body, response_headers, http_status, content_type = connection.post(url=occi_config['url'] + '/-/', headers=['Content-Type: text/plain'], body=new_mixin)
        check_create, tmp_err_msg = check_http_status("200 OK", http_status)
        err_msg += tmp_err_msg

        if not check_create:
            print >> sys.stderr, body

        return [check_create, err_msg]


def CORE_READ_URL(filter):
    err_msg = []
    check_200ok = False
    check = True

    body, response_headers, http_status, content_type, check_pretest = Test.pretest_http_status("200 OK", err_msg)

    if content_type:
        check, tmp_err_msg = check_content_type(content_type)
        err_msg += tmp_err_msg

    if not check_pretest:
        return [False, err_msg, None]

    mixin = Test.search_category(filter)
    #kind = Test.search_category({'class': 'kind'})

    if not mixin:
        err_msg.append('No required OCCI Category found')
        return [False, err_msg, None]

    for category in [mixin]:
        body, response_headers, http_status, content_type = connection.get(url=occi_config['url'] + category['location'])
        try:
            locations = renderer.parse_locations(body, response_headers)
        except occi.ParseError as pe:
            locations = []
            check = False
            err_msg.append(str(pe))

        if re.match(r'^HTTP/.* 200 OK', http_status):
            check_200ok = True
        else:
            err_msg.append('Returned HTTP status is not 200 OK (%s)' % http_status)

    return [check_pretest and check and check_200ok, err_msg, locations]


class CORE_READ001(Test):
    objective = 'Retrieve the URLs of all OCCI Entities belonging to an OCCI Kind or OCCI Mixin'

    @classmethod
    def test(self=None):
        check, err_msg, urls = CORE_READ_URL(occi_config['occi.tests.category'])
        return [check, err_msg]


def CORE_READ002_COMMON(category, links=[]):
    check_url = True
    check_200ok = False
    err_msg = []
    headers = []

    headers.append('Content-Type: text/occi')
    headers += renderer_httpheaders.render_category(occi.Category({'term': category['term'], 'scheme': category['scheme'], 'class': category['class']}))[1]

    body, response_headers, http_status, content_type = connection.get(url=occi_config['url'] + category['location'], headers=headers)
    try:
        locations = renderer.parse_locations(body, response_headers)
    except occi.ParseError as pe:
        locations = []
        check_url = False
        err_msg.append(str(pe))
    links[0:] = locations

    if re.match(r'^HTTP/.* 200 OK', http_status):
        check_200ok = True
    else:
        err_msg.append('Returned HTTP status is not 200 OK (%s)' % http_status)

    check_ct2, tmp_err_msg = check_content_type(content_type)
    err_msg += tmp_err_msg

    return [check_url and check_ct2 and check_200ok, err_msg]


class CORE_READ002(Test):
    objective = 'Retrieve the URLs of the OCCI Entities belonging to an OCCI Kind or OCCI Mixin and related to an OCCI Category filter'

    @classmethod
    def test(self=None):
        err_msg = []
        check = True

        body, response_headers, http_status, content_type, check_pretest = Test.pretest_http_status("200 OK", err_msg)

        if content_type:
            check, tmp_err_msg = check_content_type(content_type)
            err_msg += tmp_err_msg

        if not check_pretest:
            return [False, err_msg]

        kinds = Test.search_category({'class': 'kind'})
        found = False
        check_read = False
        for category in kinds:
            filter = Test.search_category({'rel': '%s%s' %
                                           ('http://schemas.ogf.org/occi/core#', 'resource')})
            if filter is None:
                continue
            else:
                found = True
                check_read, tmp_err_msg = CORE_READ002_COMMON(category=filter)

        if not found:
            err_msg += ['No desired OCCI Mixin found']

        return [check and check_read, err_msg]


def CORE_READ_DESCRIPTION(filter=None):
    """Read OCCI Entity Description

    :param string filter{}: required attribute names and values
    :return: status, err_msg, url, entity description
    :rtype: [bool, string[], string, occi.Category[], occi.Link[], occi.Attribute[]]
    """
    categories = []
    links = []
    attributes = []
    found = False
    entity_url = None

    check, err_msg, urls = CORE_READ_URL(occi_config['occi.tests.category'])
    if not check:
        return [False, err_msg, entity_url, None, None, None]

    #print urls
    for entity_url in urls:
        #print base_url, url
        body, headers, http_status, content_type = connection.get(url=entity_url)
        check1, tmp_err_msg = check_http_status("200 OK", http_status)
        err_msg += tmp_err_msg
        if content_type:
            check2, tmp_err_msg = check_requested_content_type(content_type)
            err_msg += tmp_err_msg
        if not check1 or not check2:
            return [False, err_msg, entity_url, None, None, None]

        try:
            categories, links, attributes = renderer.parse_resource(body, headers)
        except occi.ParseError as pe:
            err_msg += [str(pe)]
            return [False, err_msg, entity_url, None, None, None]
        if not categories:
            err_msg += ['HTTP Response doesn\'t contain categories']
            return [False, err_msg, entity_url, None, None, None]
        #print 'Got OCCI Entity: ' + str(categories[0])

        if match_entity(attributes, filter):
            #print 'Hit!'
            found = True
            break

    if not found:
        err_msg.append('No required OCCI Entity instance found')
    return [found, err_msg, entity_url, categories, links, attributes]


class CORE_READ007(Test):
    objective = 'Retrieve the description of an OCCI Entity'

    @classmethod
    def test(self=None):
        check, err_msg, url, categories, links, attributes = CORE_READ_DESCRIPTION({})
        return [check, err_msg]


def get_attributes(attribute_definitions, attributes, err_msg):
    """Fill attribute values from example_attributes for all required attributes.

    :param occi.AttributeDefinition attribute_definitions[]: attribute definitions
    :param occi.Attribute attributes{}: result attributes dictionary
    :param string err_msg[]: list of errors to append
    :return: all required attributes has a value
    :rtype: bool
    """
    has_all_attributes = True

    #print 'list of attributes: %s' % attributes
    for ad in attribute_definitions:
        #print 'attribute: %s' % ad
        if ad.isrequired() and not ad.isimmutable() and not ad['name'] in attributes:
            if ad['name'] not in example_attributes:
                err_msg.append('Tests error: unknown attribute %s' % ad['name'])
                has_all_attributes = False
                continue
            #print 'adding attribute: %s' % ad
            attributes[ad['name']] = example_attributes[ad['name']]

    return has_all_attributes


class CORE_DELETE001(Test):
    objective = 'Delete an OCCI Entity'

    @classmethod
    def test(self=None):
        err_msg = []

        body, response_headers, http_status, content_type, check_pretest = Test.pretest_http_status("200 OK", err_msg)

        category = occi.Category(occi_config['occi.tests.category'])
        check, err_msg = INFRA_CREATE_COMMON('compute', [category], [], err_msg)
        if not check:
            return [False, err_msg]

        check, err_msg, tmp_urls = CORE_READ_URL(occi_config['occi.tests.category'])

        if not tmp_urls:
            err_msg += ["OCCI entity URL not found!"]
            return [False, err_msg]

        url = tmp_urls[0]
        if occi_config['curlverbose']:
            print '[OCCI/CORE/DELETE/001] entity URL: %s' % str(url)

        body, response_headers, http_status, content_type = connection.get(url=url)
        check_exist1, tmp_err_msg = check_http_status("200 OK", http_status)
        err_msg += tmp_err_msg
        if occi_config['curlverbose']:
            print '[OCCI/CORE/DELETE/001] entity body: %s' % str(body)

        body, response_headers, http_status, content_type = connection.delete(url=url)
        check_delete1, tmp_err_msg = check_http_status('2\d\d ', http_status)
        err_msg += tmp_err_msg

        # It takes some time to delete machine, second delete action force it
        # Not testing result of the operation (various backends have different behaviour)
        body, response_headers, http_status, content_type = connection.delete(url=url)

        body, response_headers, http_status, content_type = connection.get(url=url)
        check_exist2, tmp_err_msg = check_http_status("404 Not Found", http_status)
        err_msg += tmp_err_msg

        Test.clear_categories()

        return [check_exist1 and check_exist2 and check_delete1, err_msg]


class CORE_UPDATE001(Test):
    """Full update of a specific OCCI Entity

    Requires existing compute machine.

    OpenNebula issues:

    * https://github.com/EGI-FCTF/rOCCI-server/issues/125: poweroff state required
    * https://github.com/EGI-FCTF/rOCCI-server/issues/126: not all attributes implemented
    * https://github.com/EGI-FCTF/rOCCI-server/issues/128: parse error
    """

    objective = 'Full update of a specific OCCI Entity'

    @classmethod
    def test(self=None):
        err_msg = []
        check_response = True

        check, err_msg, urls = CORE_READ_URL(occi_config['occi.tests.category'])
        if not urls:
            err_msg.append('No OCCI Entity instance found')
            return [False, err_msg]
        url = urls[0]
        if occi_config['curlverbose']:
            print '[OCCI/CORE/UPDATE/001] all entities: %s' % str(urls)
            print '[OCCI/CORE/UPDATE/001] selected entity: %s' % str(url)

        body, response_headers, http_status, content_type = connection.get(url=url)

        categories, links, attributes = renderer.parse_resource(body, response_headers)
        if occi_config['curlverbose']:
            print '[OCCI/CORE/UPDATE/001] resource categories: %s' % str(categories)
            print '[OCCI/CORE/UPDATE/001] resource links: %s' % str(links)
            print '[OCCI/CORE/UPDATE/001] resource attributes: %s' % str(attributes)

        # change one attribute
        a = None
        for a in attributes:
            if a['name'] == 'occi.core.title':
                break
        if a is None or a['name'] != 'occi.core.title':
            a = occi.Attribute({'name': 'occi.core.title', 'value': gen_id('c_pOCCI')})
            attributes.append(a)
        else:
            a['value'] = gen_id(a['value'])
        body, headers = renderer.render_resource(categories, links, attributes)

        # update
        body, response_headers, http_status, content_type = connection.put(url=url, headers=['Content-Type: %s' % occi_config['mimetype']] + headers, body=body)
        #print body
        #print http_status
        Test.clear_categories()

        check_ct, tmp_err_msg = check_content_type(content_type)
        err_msg += tmp_err_msg
        if content_type != occi_config['mimetype']:
            err_msg += ['Content-Type is not requested "%s"' % occi_config['mimetype']]
            check_ct = False

        check, tmp_err_msg = check_http_status("200 OK", http_status)
        if check:
            # response contains OCCI Entity description
            response_categories, response_links, response_attributes = renderer.parse_resource(body, response_headers)
            if not response_categories or not response_attributes:
                err_msg += ['HTTP Response doesn\'t contain categories or attributes']
                check_response = False
        else:
            check, tmp_err_msg = check_http_status("201 Created", http_status)
            if check:
                response_urls = renderer_httpheaders.parse_locations(None, response_headers)
                #print 'Entity URL'
                #print 'Response URLs:\n  '
                #print response_urls
                check_response = False
                for url in response_urls:
                    if response_url == url:
                        check_response = True
                if not check_response:
                    err_msg += ['HTTP Location headers is not OCCI Entity URL (%s)' % response_url]

        if not check:
            err_msg.append('HTTP status is neither 200 OK nor 201 Created (%s)' % http_status)
            print >> sys.stderr, body

        return [check and check_ct and check_response, err_msg]


class CORE_MISC001(Test):
    objective = 'Trigger OCCI Action on existing OCCI Entity'

    @classmethod
    def test(self=None):
        err_msg = []
        actions = set()

        check, err_msg, entity_url, categories, links, attributes = CORE_READ_DESCRIPTION(filter=occi_config['occi.tests.entity'])
        if not check:
            return [False, err_msg]

        for cat in categories:
            model_cat = Test.get_category(cat['scheme'] + cat['term'])
            #print model_cat['term']
            if 'actions' in model_cat:
                #print model_cat['actions']
                actions |= set(model_cat['actions'])
        if occi_config['curlverbose']:
            print '[OCCI/CORE/MISC/001] actions: ' + str(list(actions))

        # select appropriate action, fallback to the first one
        action = list(actions)[0]
        action_start = action
        action_stop = action
        for act in actions:
            if re.search('#start$', act):
                action_start = act
            if re.search('#suspend$', act):
                action_stop = act
        for a in attributes:
            if a['name'] == 'occi.compute.state':
                break
        if a['name'] == 'occi.compute.state':
            if occi_config['curlverbose']:
                print '[OCCI/CORE/MISC/001] state: %s' % a['value']
            if a['value'] == 'active':
                action = action_stop
            else:
                action = action_start
        if occi_config['curlverbose']:
            print '[OCCI/CORE/MISC/001] selected action: %s' % action

        action_cat = Test.get_category(action)
        if not action_cat:
            return [False, 'Action "%s" not found in OCCI Model' % action]
        #print action_cat

        url = entity_url
        url += '?action=' + action_cat['term']
        if occi_config['curlverbose']:
            print '[OCCI/CORE/MISC/001] calling action: %s' % url

        body, headers = renderer.render_category(action_cat)

        body, response_request, http_status, content_type = connection.post(url=url, headers=['Content-Type: %s' % occi_config['mimetype']] + headers, body=body)

        check, tmp_err_msg = check_http_status("200 OK", http_status)
        err_msg += tmp_err_msg

        return [check, err_msg]


def INFRA_CREATE_COMMON(resource, categories, additional_attributes, err_msg):
    """Generic help function to create OCCI Infrastructure resources.

    HTTP Headers renderer is always used.

    :param string resource: OCCI Category term (compute, storage, network)
    :param occi.Category categories[]: OCCI Categories to add to rendering
    :param occi.AttributeDefinition additional_attributes[]: additional attributes to set from example defaults
    :param string err_msg[]: error messages list to append
    :return: status and error message list
    :rtype: [bool, string[]]
    """
    has_kind = True
    has_all_attributes = True
    all_attributes = []
    attributes = {}

    kind = Test.search_category({'class': 'kind', 'term': resource, 'scheme': 'http://schemas.ogf.org/occi/infrastructure#'})
    #print kind

    if not kind:
        has_kind = False
        err_msg.append('No OCCI Kind found')
    if not kind:
        return [False, err_msg]

    for item in ['location', 'term', 'scheme']:
        if item not in kind.keys():
            has_kind = False
            err_msg.append('No %s in OCCI Kind' % item)

    if 'attributes' in kind:
        all_attributes += kind['attributes']

    if additional_attributes is not None:
        all_attributes += additional_attributes

    #print 'list of attributes: %s' % repr(all_attributes)
    if not get_attributes(all_attributes, attributes, err_msg):
        has_all_attributes = False
    #print 'list of result attribute keys: %s' % repr(attributes.keys())

    new_cat_s, new_cat_h = renderer.render_resource(categories, None, attributes.values())

    body, response_request, http_status, content_type = connection.post(url=occi_config['url'] + kind['location'], headers=['Content-Type: %s' % occi_config['mimetype']] + new_cat_h, body=new_cat_s)
    check_create, tmp_err_msg = check_http_status("(200 OK|201 Created)", http_status)
    err_msg += tmp_err_msg

    check_ct, tmp_err_msg = check_content_type(content_type)
    err_msg += tmp_err_msg

    entities = check_body_entities(body, response_request, err_msg)
    check_br = (entities is not None and entities)

    check_rct, tmp_err_msg = check_requested_content_type(content_type)
    err_msg += tmp_err_msg

    if not check_create:
        print >> sys.stderr, body

    Test.clear_categories()

    body, response_headers, http_status, content_type = connection.get(url=occi_config['url'] + kind['location'])
    entities2 = check_body_entities(body, response_headers, err_msg)
    # check if the entity is really created (just check the first: entities[0])
    check_created = False
    if entities:
        for line in entities2:
            if line == entities[0]:
                check_created = True
                break
    if not check_created:
        err_msg.append('OCCI %s Resource hasn\'t been successfully created' % resource.title())

    return [has_kind and has_all_attributes and check_create and check_created and check_ct and check_br and check_rct and check_created, err_msg]


class INFRA_CREATE001(Test):
    objective = 'Create an OCCI Compute Resource'

    @classmethod
    def test(self=None):
        err_msg = []
        category = occi.Category({'term': 'compute', 'scheme': 'http://schemas.ogf.org/occi/infrastructure#', 'class': 'kind'})

        body, response_headers, http_status, content_type, check_pretest = Test.pretest_http_status("200 OK", err_msg)
        if not check_pretest:
            return [False, err_msg]

        return INFRA_CREATE_COMMON('compute', [category], [], err_msg)


class INFRA_CREATE002(Test):
    objective = 'Create an OCCI Storage Resource'

    @classmethod
    def test(self=None):
        err_msg = []
        category = occi.Category({'term': 'storage', 'scheme': 'http://schemas.ogf.org/occi/infrastructure#', 'class': 'kind'})
        additional_attributes = [
            occi.AttributeDefinition({"name": "occi.core.title", "required": True}),
            occi.AttributeDefinition({"name": "occi.storage.size", "required": True}),
        ]

        body, response_headers, http_status, content_type, check_pretest = Test.pretest_http_status("200 OK", err_msg)
        if not check_pretest:
            return [False, err_msg]

        return INFRA_CREATE_COMMON('storage', [category], additional_attributes, err_msg)


class INFRA_CREATE003(Test):
    objective = 'Create an OCCI Network Resource'

    @classmethod
    def test(self=None):
        err_msg = []
        category = occi.Category({'term': 'network', 'scheme': 'http://schemas.ogf.org/occi/infrastructure#', 'class': 'kind'})
        additional_attributes = [occi.AttributeDefinition({"name": "occi.core.title", "required": True})]

        body, response_headers, http_status, content_type, check_pretest = Test.pretest_http_status("200 OK", err_msg)
        if not check_pretest:
            return [False, err_msg]

        return INFRA_CREATE_COMMON('network', [category], additional_attributes, err_msg)


class INFRA_CREATE004(Test):
    objective = 'Create an OCCI Compute Resource using an OS and resource template'

    @classmethod
    def test(self=None):
        err_msg = []
        categories = [
            occi.Category({'term': 'compute', 'scheme': 'http://schemas.ogf.org/occi/infrastructure#', 'class': 'kind'})
        ]

        body, response_headers, http_status, content_type, check_pretest = Test.pretest_http_status("200 OK", err_msg)
        if not check_pretest:
            return [False, err_msg]

        # OS template
        os_tpl = Test.search_category({'class': 'mixin', 'rel': 'http://schemas.ogf.org/occi/infrastructure#os_tpl'})
        if occi_config['curlverbose']:
            print '[OCCI/INFRA/CREATE/004] os_tpl: %s' % str(os_tpl)
        if not os_tpl:
            err_msg.append('No OCCI OS Template found')
            return [False, err_msg]

        # 'term': 'uuid_ttylinux_0', 'scheme': 'http://occi.myriad5.zcu.cz/occi/infrastructure/os_tpl#', 'class': 'mixin'
        categories.append(occi.Category({'term': os_tpl['term'], 'scheme': os_tpl['scheme'], 'class': 'mixin'}))

        if 'attributes' in os_tpl:
            tpl_attributes = os_tpl['attributes']
        else:
            tpl_attributes = []

        # resource template
        resource_tpl = Test.search_category({'class': 'mixin', 'rel': 'http://schemas.ogf.org/occi/infrastructure#resource_tpl'})
        if occi_config['curlverbose']:
            print '[OCCI/INFRA/CREATE/004] resource_tpl: %s' % str(resource_tpl)
        if not resource_tpl:
            err_msg.append('No OCCI Resource Template found')
            return [False, err_msg]
        if resource_tpl:
            categories.append(occi.Category({'term': resource_tpl['term'], 'scheme': resource_tpl['scheme'], 'class': 'mixin'}))
            if 'attributes' in resource_tpl:
                tpl_attributes += resource_tpl['attributes']

        if occi_config['curlverbose']:
            print '[OCCI/INFRA/CREATE/004] attributes: %s' % str(tpl_attributes)

        return INFRA_CREATE_COMMON('compute', categories, tpl_attributes, err_msg)


class INFRA_CREATE005(Test):
    """
    Unsupported test, os_tpl required.

    It can be called by pOCCI -t 'INFRA/CREATE/005'
    """

    objective = 'Create an OCCI Compute Resource with an OCCI Storagelink and an OCCI Networkinterface'

    @classmethod
    def test(self=None):
        network_links = []
        storage_links = []
        err_msg = []
        check = True

        body, response_headers, http_status, content_type, check_pretest = Test.pretest_http_status("200 OK", err_msg)

        if content_type:
            check, tmp_err_msg = check_content_type(content_type)
            err_msg += tmp_err_msg

        if not check_pretest:
            return [False, err_msg]

        storage = Test.search_category({'term': 'storage', 'scheme': 'http://schemas.ogf.org/occi/infrastructure#'})
        network = Test.search_category({'term': 'network', 'scheme': 'http://schemas.ogf.org/occi/infrastructure#'})
        check_read, tmp_err_msg = CORE_READ002_COMMON(category=storage, links=storage_links)
        if not check_read:
            check = False
        err_msg += tmp_err_msg

        check_read, tmp_err_msg = CORE_READ002_COMMON(category=network, links=network_links)
        if not check_read:
            check = False
        err_msg += tmp_err_msg

        print storage_links
        print network_links
        if not storage_links or not network_links:
            if not storage_links:
                err_msg.append('No storage found')
            if not network_links:
                err_msg.append('No network found')
            return [False, err_msg]

        compute = Test.search_category({'term': 'compute', 'scheme': 'http://schemas.ogf.org/occi/infrastructure#'})

        new_compute = 'Category: %s; scheme="%s"; class="%s"\n\r\
Link: <%s>; rel="%s"; category="%s"\n\r\
Link: <%s>; rel="%s"; category="%s"\n\r\
' % (compute['term'], compute['scheme'], compute['class'], storage_links[0], storage['scheme'] + storage['term'], 'http://schemas.ogf.org/occi/infrastructure#storagelink', network_links[0], network['scheme'] + network['term'], 'http://schemas.ogf.org/occi/infrastructure#networkinterface')

        body, response_headers, http_status, content_type = connection.post(url=occi_config['url'] + compute['location'], headers=['Content-Type: text/plain'], body=new_compute)
        check_create, tmp_err_msg = check_http_status("201 Created", http_status)
        err_msg += tmp_err_msg

        if not check_create:
            print >> sys.stderr, body

        Test.clear_categories()

        return [check and check_create, err_msg]


def INFRA_CREATE_LINK(resource_name, resource_type):
    """
    Opennebula requires running compute instance.

    :param string resource_name: Resource Name (storage, network)
    :param string resource_type: Resource Type (link, interface)
    :return: status, err_msg
    :rtype: [bool, string[]]
    """

    err_msg = []
    check = True
    check_link = False
    compute_links = []
    resource_links = []
    resourcelink = None
    attributes = {}
    attribute_definitions = []

    body, response_headers, http_status, content_type, check_pretest = Test.pretest_http_status("200 OK", err_msg)
    if not check_pretest:
        return [False, err_msg]

    compute = Test.search_category({'term': 'compute', 'scheme': 'http://schemas.ogf.org/occi/infrastructure#'})
    resource = Test.search_category({'term': resource_name, 'scheme': 'http://schemas.ogf.org/occi/infrastructure#'})

    check_read, tmp_err_msg = CORE_READ002_COMMON(category=compute, links=compute_links)
    if not check_read:
        check = False
    err_msg += tmp_err_msg

    check_read, tmp_err_msg = CORE_READ002_COMMON(category=resource, links=resource_links)
    if not check_read:
        check = False
    err_msg += tmp_err_msg

    #print resource_links
    #print compute_links

    if not resource_links or not compute_links:
        if not resource_links:
            err_msg.append('No %s found' % resource_name)
        if not compute_links:
            err_msg.append('No compute found')
        return [False, err_msg]

    resourcelink = Test.search_category({'term': '%s%s' % (resource_name, resource_type), 'scheme': 'http://schemas.ogf.org/occi/infrastructure#', 'class': 'kind'})
    if not resourcelink:
        err_msg.append('No %slink kind found' % resource_name)
        return [False, err_msg]
    #print resourcelink

    if 'attributes' in resourcelink:
        attribute_definitions += resourcelink['attributes']

    attributes['occi.core.id'] = occi.Attribute({'name': 'occi.core.id', 'value': gen_id('%s%s' % (resource_name.capitalize(), resource_type))})
    attributes['occi.core.source'] = occi.Attribute({'name': 'occi.core.source', 'value': compute_links[0]})
    attributes['occi.core.target'] = occi.Attribute({'name': 'occi.core.target', 'value': resource_links[0]})
    if not get_attributes(attribute_definitions, attributes, err_msg):
        check = False
    #print attributes

    new_resourcelink_s, new_resourcelink_h = renderer.render_resource(
        categories=[
            occi.Category({'term': resourcelink['term'], 'scheme': resourcelink['scheme'], 'class': resourcelink['class']}),
        ],
        links=None,
        attributes=attributes.values()
    )

    body, response_headers, http_status, content_type = connection.post(url=occi_config['url'] + resourcelink['location'], headers=['Content-Type: %s' % occi_config['mimetype']] + new_resourcelink_h, body=new_resourcelink_s)

    check_create, tmp_err_msg = check_http_status("201 Created", http_status)
    err_msg += tmp_err_msg

    if not check_create:
        print >> sys.stderr, body

    Test.clear_categories()

    resource_link = resource_links[0]
    resource_link_rel = urlparse.urlparse(resource_link).path

    body, response_headers, http_status, content_type = connection.get(url=compute_links[0])

    result_links = []
    try:
        result_categories, result_links, result_attributes = renderer.parse_resource(body, response_headers)
    except occi.ParseError as pe:
        err_msg += [str(pe)]

    #print resource_link
    #print result_links
    for l in result_links:
        # accept both: relative and absolute URI
        if l['uri'] == resource_link or l['uri'] == resource_link_rel:
            check_link = True
            break

    if not check_link:
        err_msg += ["OCCI Compute Resource is NOT linked with OCCI %s Resource!" % (resource_name.capitalize())]
        print >> sys.stderr, body

    return [check and check_create and check_link, err_msg]


class INFRA_CREATE006(Test):
    """
    Opennebula requires running compute instance.
    """

    objective = 'Create an OCCI Storagelink between an existing OCCI Compute and OCCI Storage Resource'

    @classmethod
    def test(self=None):
        return INFRA_CREATE_LINK('storage', 'link')


class INFRA_CREATE007(Test):
    """
    Opennebula requires running compute instance.
    """

    objective = 'Create an OCCI Networkinterface between an existing OCCI Compute and OCCI Network Resource'

    @classmethod
    def test(self=None):
        return INFRA_CREATE_LINK('network', 'interface')



#!/usr/bin/python

import getopt
import os
import re
import sys
import version

import occi
import render


inputmime = 'text/plain'
outputmime = 'text/plain'
messagetype = 'categories'

messagetypes = ['categories', 'entities', 'resource']


def usage(name=__file__):
    print '%s [OPTIONS]\n\
\n\
OPTIONS:\n\
  -h, --help\n\
  -i, --input-mime MIME-TYPE .... input mime-type [text/plain]\n\
  -o, --output-mime MIME-TYPE ... output mime-type [text/plain]\n\
  -t, --type OCCI-TYPE .......... OCCI message type [categories]\n\
  -V, --version ................. print version information\n\
\n\
MIME-TYPE: text/plain, text/occi, text/uri-list\n\
OCCI-TYPE: %s\n\
' % (os.path.basename(name), ', '.join(messagetypes))


def read_input(strip=False):
    for line in sys.stdin:
        if strip:
            yield line.rstrip('\n\r')
        else:
            yield line


def main(argv=sys.argv[1:]):
    global inputmime, outputmime, messagetype

    occi_parser = None
    occi_renderer = None

    try:
        opts, args = getopt.getopt(argv, 'hi:o:t:V', ['--help', '--input-mime=', '--output-mime=', 'type=', 'version'])
    except getopt.GetoptError:
        usage()
        sys.exit(2)
    for opt, arg in opts:
        if opt in ['-h', '--help']:
            usage()
            sys.exit()
        elif opt in ['-i', '--input-mime']:
            inputmime = arg
        elif opt in ['-o', '--output-mime']:
            outputmime = arg
        elif opt in ['-t', '--type']:
            messagetype = arg
        elif opt in ["-V", "--version"]:
            print version.__version__
            sys.exit()

    if messagetype not in messagetypes:
        print >> sys.stderr, 'OCCI message type expected: %s' % ', '.join(messagetypes)
        sys.exit(2)

    occi_parser = render.create_renderer(inputmime)
    if not occi_parser:
        print >> sys.stderr, 'OCCI parser can\'t be initialized (wrong mime-type "%s"?)' % inputmime
        sys.exit(2)

    occi_renderer = render.create_renderer(outputmime)
    if not occi_renderer:
        print >> sys.stderr, 'OCCI renderer can\'t be initialized (wrong mime-type "%s"?)' % outputmime
        sys.exit(2)

    if re.match(r'text/occi(;.*)?$', inputmime):
        body = None
        headers = list(read_input(strip=False))
    else:
        body = list(read_input(strip=True))
        headers = None

    try:
        if messagetype in ['categories']:
            categories = occi_parser.parse_categories(body, headers)
            [body, headers] = occi_renderer.render_categories(categories)
        elif messagetype in ['entities']:
            urls = occi_parser.parse_locations(body, headers)
            [body, headers] = occi_renderer.render_locations(urls)
        elif messagetype in ['resource']:
            [categories, links, attributes] = occi_parser.parse_resource(body, headers)
            [body, headers] = occi_renderer.render_resource(categories, links, attributes)
    except occi.ParseError as perr:
        print >> sys.stderr, str(perr)
        sys.exit(1)
    except occi.RenderError as rerr:
        print >> sys.stderr, str(rerr)
        sys.exit(1)

    if body:
        sys.stdout.write(body)
    if headers:
        sys.stdout.write('\n'.join(headers) + '\n')

if __name__ == '__main__':
    main(sys.argv[1:])



import collections
import re

import occi
from render_base import Renderer, check_url


eol = '\r\n'


def text_attribute_def(ad=None):
    s = ad['name']
    immutable = ('immutable' in ad) and ad['immutable']
    required = ('required' in ad) and ad['required']
    if immutable and required:
        s += '{immutable required}'
    elif immutable and not required:
        s += '{immutable}'
    elif not immutable and required:
        s += '{required}'
    return s


def text_attribute_defs(ads=None):
    text_ads = []
    if ads:
        for ad in ads:
            text_ads.append(text_attribute_def(ad))
    return ' '.join(text_ads)


def text_actions(actions=None):
    if actions:
        return ' '.join(actions)
    else:
        return None


def text_category(category=None):
    s = '%s;scheme="%s";class="%s"' % (category['term'], category['scheme'], category['class'])

    for item in ['title', 'rel', 'location']:
        if item in category:
            s += ';%s="%s"' % (item, category[item])

    if 'attributes' in category:
        s += ';%s="%s"' % ('attributes', text_attribute_defs(category['attributes']))

    if 'actions' in category:
        s += ';%s="%s"' % ('actions', text_actions(category['actions']))

    return s


def text_attribute_value(attribute):
    """Render OCCI Attribute value.

    :param occi.Attribute attribute: attribute with a value to render
    :return: attribute value rendering
    :rtype: string
    """

    if 'type' in attribute:
        type = attribute['type']
    else:
        type = 'string'
    value = attribute['value']

    if type in ['string', 'enum']:
        return '"' + value + '"'
    elif type == 'number':
        return repr(value)
    elif type == 'bool':
        if value:
            return "true"
        else:
            return "false"


def text_attribute_repr(attribute):
    """Render one OCCI Attribute.
    """

    return attribute['name'] + '=' + text_attribute_value(attribute)


def text_link_attribute(key, value):
    """Render Link Attribute
    """
    # term or quoted string, using only the quotes now
    return key + '=' + '"' + value + '"'


def text_link(link):
    s = '<%s>;rel="%s"' % (link['uri'], ' '.join(link['rel']))

    if 'self' in link:
        s += ';self="%s"' % link['self']
    if 'category' in link:
        s += ';category="%s"' % ' '.join(link['category'])

    if 'attributes' in link:
        for key, value in link['attributes'].iteritems():
            s += ';%s' % text_link_attribute(key, value)

    return s


class TextRenderer(Renderer):
    """Plain Text OCCI Renderer

    Empty array is always returned as headers during rendering.
    """

    reChunks = re.compile(r';\s*')
    reCategory = re.compile(r'^(?i)category:\s*')
    reLink = re.compile(r'^(?i)link:\s*')
    reAttribute = re.compile(r'^(?i)x-occi-attribute:\s*')
    reKeyValue = re.compile(r'\s*?=\s*')
    reKeyCheck = re.compile(r'[A-Za-z0-9_\.-]*$')
    reQuoted = re.compile(r'^"(.*)"$')
    reSP = re.compile(r'\s+')
    reAttributes = re.compile(r'([^\{ ]+)(\{[^\}]*\})?\s*')
    reLocation = re.compile(r'^((?i)x-occi-location|location):\s*(.*)')
    reQuotedLink = re.compile(r'^<(.*)>$')
    reStringUnescape = re.compile(r'\\(.)')
    reNumber = re.compile(r'^([0-9\.+-]+)$')
    reIntNumber = re.compile(r'^([0-9+-]+)$')
    reBool = re.compile(r'^(true|false)$')


    def render_category(self, category):
        """Render OCCI Category

        :param occi.Category category: OCCI Category object
        :return: render result
        :rtype: [string, string[]]
        """
        return ['Category: ' + text_category(category), []]


    def render_categories(self, categories):
        """Render OCCI Category collection

        :param occi.Category category[]: OCCI Category array
        :return: render result
        :rtype: [string, string[]]
        """
        res = []
        for category in categories:
            cat_s, cat_h = self.render_category(category)
            res.append(cat_s)
        return [eol.join(res) + eol, []]


    def render_resource(self, categories, links=None, attributes=None):
        """Render OCCI Resource instance

        :param occi.Category category: OCCI Category object
        :param occi.Link links[]: OCCI Link array
        :param occi.Attribute attributes[]: OCCI Attribute array
        :return: render result
        :rtype: [string, string[]]
        :return: render result
        """
        Renderer.render_resource(self, categories, links, attributes)
        cat_s, cat_h = self.render_categories(categories)
        res = []
        if links is not None:
            for link in links:
                res.append(self.render_link(link))
        if attributes is not None:
            for attr in attributes:
                res.append(self.render_attribute(attr))
        if res:
            return [cat_s + eol.join(res) + eol, []]
        else:
            return [cat_s, []]


    def render_link(self, link):
        """ Render OCCI Link

        :param occi.Link link: OCCI Link object
        :return: render result
        :rtype: string
        """
        return 'Link: ' + text_link(link)


    def render_attribute(self, attribute):
        """ Render Attribute

        :param occi.Attribute attribute: OCCI Attribute object
        :return: render result
        :rtype: string
        """
        return 'X-OCCI-Attribute: ' + text_attribute_repr(attribute)


    def render_locations(self, locations):
        """ Render Locations

        :param string location[]: location URI
        :return: render result
        :rtype: [string, string[]]
        """
        if not locations:
            return ''
        s = []
        for location in locations:
            s.append('X-OCCI-Location: ' + location + eol)
        return [''.join(s), []]


    def parse_attribute_defs(self, body):
        """ Parse OCCI Attribute Definitions.

        Example::

           occi.core.id{immutable required} occi.core.title occi.core.target occi.core.source{required}

        :param string body: text to parse
        :return: array of OCCI Attribute Definition
        :rtype: occi.AttributeDefinition[]
        """
        result = []

        m = True
        while m:
            m = TextRenderer.reAttributes.match(body)
            if not m:
                break
            matches = m.groups()
            name = matches[0]
            attrs = matches[1]
            body = body[m.end():]

            if attrs:
                attrs = attrs[1:-1]
                attrs = TextRenderer.reSP.split(attrs)

            attribute = occi.AttributeDefinition({'name': name})
            if attrs:
                for a in attrs:
                    if a == 'required':
                        attribute['required'] = True
                    elif a == 'immutable':
                        attribute['immutable'] = True
                    else:
                        raise occi.ParseError('Unknown field in OCCI attribute definitions', a)
            result.append(attribute)

        if body:
            raise occi.ParseError('Error parsing OCCI attribute definitions', body)

        return result


    def parse_actions(self, body):
        """Parse OCCI Actions.

        Example::

           http://schemas.ogf.org/occi/infrastructure/compute/action#start http://schemas.ogf.org/occi/infrastructure/compute/action#stop http://schemas.ogf.org/occi/infrastructure/compute/action#restart http://schemas.ogf.org/occi/infrastructure/compute/action#suspend

        :param string body: text to parse
        :return: array of string
        :rtype: string[]
        """
        actions = TextRenderer.reSP.split(body)
        for action in actions:
            # let's require scheme and hostname in scheme URI
            if not check_url(action, scheme=True, host=True):
                raise occi.ParseError('URI expected as an action', action)
        return actions


    def parse_category_body(self, body):
        """Parse OCCI Category body

        Example::

           entity;scheme="http://schemas.ogf.org/occi/core#";class="kind";title="entity";location="/entity/";attributes="occi.core.id{immutable required} occi.core.title"

        :param string body: text to parse
        :return: OCCI Category
        :rtype: occi.Category
        """
        category = occi.Category()

        chunks = TextRenderer.reChunks.split(body)

        if not chunks[0]:
            raise occi.ParseError('Invalid format of category, term expected', body)

        category['term'] = chunks[0]

        # skip the first chunk (category term)
        for chunk in chunks[1:]:
            keyvalue = TextRenderer.reKeyValue.split(chunk, 1)
            if len(keyvalue) != 2:
                raise occi.ParseError('Key/value pair expected in category', chunk)

            key = keyvalue[0]
            value = keyvalue[1]
            keymatch = TextRenderer.reKeyCheck.match(key)
            if keymatch is None:
                raise occi.ParseError('Invalid characters in category property', chunk)
            # every value quoted, only class has quoting optional
            valuematch = TextRenderer.reQuoted.match(value)
            if valuematch is None and key != 'class':
                raise occi.ParseError('Category value not properly quoted or unexpected EOF', chunk)
            if valuematch:
                value = valuematch.group(1)
            # sanity check: there should not be any quotes now
            if value[0] == '"' or (len(value) >= 2 and value[-1] == '"'):
                raise occi.ParseError('Unexpected quotes in category', chunk)

            if key == 'location':
                if not check_url(value):
                    raise occi.ParseError('URL is not valid in OCCI Category location', chunk)
                category[key] = value
            elif key == 'scheme':
                if not check_url(value):
                    raise occi.ParseError('URL is not valid in OCCI Category scheme', chunk)
                category[key] = value
            elif key == 'attributes':
                category[key] = self.parse_attribute_defs(value)
            elif key == 'actions':
                category[key] = self.parse_actions(value)
            elif key in ['class', 'title', 'rel']:
                category[key] = value
            else:
                raise occi.parseerror('unknown key "%s" in category' % key, chunk)

        if not category.validate():
            raise occi.ParseError('Missing fields in OCCI Category', body)

        return category


    def parse_link_body(self, body):
        """Parse OCCI Link body

        Example::

           </storage/0>;rel="http://schemas.ogf.org/occi/infrastructure#storage";self="/link/storagelink/compute_103_disk_0";category="http://schemas.ogf.org/occi/infrastructure#storagelink http://opennebula.org/occi/infrastructure#storagelink";occi.core.id="compute_103_disk_0";occi.core.title="ttylinux";occi.core.target="/storage/0";occi.core.source="/compute/103";occi.storagelink.deviceid="/dev/hda";occi.storagelink.state="active"

        :param string body: text to parse
        :return: OCCI Link
        :rtype: occi.Link
        """
        link = occi.Link()

        chunks = TextRenderer.reChunks.split(body)

        if not chunks[0]:
            raise occi.ParseError('Invalid format of OCCI Link, URI and "rel" expected', body)

        matched = TextRenderer.reQuotedLink.match(chunks[0])
        if not matched:
            raise occi.ParseError('URI is not properly quoted in OCCI Link', body)

        link['uri'] = matched.group(1)
        if not check_url(link['uri']):
            raise occi.ParseError('URL is not valid in OCCI Link', link['uri'])

        # skip the first chunk (URI)
        for chunk in chunks[1:]:
            keyvalue = TextRenderer.reKeyValue.split(chunk, 1)

            key = keyvalue[0]
            value = keyvalue[1]
            keymatch = TextRenderer.reKeyCheck.match(key)
            if keymatch is None:
                raise occi.ParseError('Invalid characters in link property', chunk)
            valuematch = TextRenderer.reQuoted.match(value)
            # mandatory quoting
            if key in ['rel', 'self', 'category']:
                if valuematch is None:
                    raise occi.ParseError('Link value not properly quoted or unexpected EOF', chunk)
            # quoting of the other attributes optional
            if valuematch is not None:
                value = valuematch.group(1)
            # sanity check: there should not be any quotes now
            if value[0] == '"' or (len(value) >= 2 and value[-1] == '"'):
                raise occi.ParseError('Unexpected quotes in OCCI Link values', chunk)

            if key == 'scheme':
                if not check_url(value):
                    raise occi.ParseError('URL is not valid in OCCI Category scheme', chunk)
                link[key] = value
            elif key in ['rel', 'category']:
                link[key] = TextRenderer.reSP.split(value)
            elif key in ['self']:
                link[key] = value
            else:
                if 'attributes' not in link:
                    link['attributes'] = collections.OrderedDict()
                link['attributes'][key] = value

        if not link.validate():
            raise occi.ParseError('Missing fields in OCCI Link', body)

        return link


    def parse_attribute_value(self, body):
        """Parse OCCI Attribute value and detect its type

        string, number, and boolean types are detected, enum is returned as string.

        :param string body: text to parse
        :return: attribute type and value
        :rtype: [string, any]
        """
        if not body:
            raise occi.ParseError('OCCI Attribute value expected')

        matched = TextRenderer.reQuoted.match(body)
        if matched is not None:
            t = 'string'
            value = matched.group(1)
            value = TextRenderer.reStringUnescape.sub(r'\1', value)
            if len(value) + 2 < len(body):
                raise occi.ParseError('Unexpected quotes in OCCI Attribute value', body)
            return [t, value]

        matched = TextRenderer.reNumber.match(body)
        if matched is not None:
            t = 'number'
            if TextRenderer.reIntNumber.match(body) is not None:
                value = int(matched.group(1))
            else:
                value = float(matched.group(1))
            return [t, value]

        matched = TextRenderer.reBool.match(body)
        if matched is not None:
            t = 'boolean'
            if matched.group(1) == 'false':
                value = False
            else:
                value = True
            return [t, value]

        raise occi.ParseError('Unexpected format of OCCI Attribute value', body)


    def parse_attribute_body(self, body):
        """Parse OCCI Attribute body

        :param string body: text to parse
        :return: attribute type and value
        :rtype: occi.Attribute
        """

        keyvalue = TextRenderer.reKeyValue.split(body, 1)
        if len(keyvalue) != 2:
            raise occi.ParseError('Attribute invalid syntax', body)

        key = keyvalue[0]
        value = keyvalue[1]
        keymatch = TextRenderer.reKeyCheck.match(key)
        if keymatch is None:
            raise occi.ParseError('Invalid characters in attribute name', key)
        t, v = self.parse_attribute_value(value)

        return occi.Attribute({'name': key, 'type': t, 'value': v})


    def parse_categories(self, body, headers):
        """Parse OCCI Category Collection

        :param string body[]: text to parse
        :param string headers[]: headers to parse (unused in plain/text)
        :return: Array of OCCI Categories
        :rtype: occi.Category[]
        """
        categories = []
        category_ids = set()

        for line in body:
            if not line.strip():
                continue

            matched = TextRenderer.reCategory.match(line)
            if not matched:
                raise occi.ParseError('"category" expected', line)

            category = self.parse_category_body(line[matched.end():])

            # check uniqueness
            key = category['term'] + category['scheme']
            if key in category_ids:
                raise occi.ParseError('Category not unique (term "%s", scheme "%s")' % (category['term'], category['scheme']), line)
            category_ids.add(key)

            categories.append(category)

        return categories


    def parse_locations(self, body, headers):
        """Parse OCCI Entity collection

        :param string body[]: text to parse
        :param string headers[]: headers to parse (unused in text/plain)
        :return: Array of links
        :rtype: string[]
        """
        locations = []
        for line in body:
            if not line.strip():
                continue

            matched = TextRenderer.reLocation.match(line)
            if not matched:
                raise occi.ParseError('OCCI Location expected in OCCI Entity collection', line)
            uri = matched.group(2)
            if not check_url(uri, scheme=True, host=True):
                raise occi.ParseError('Invalid URI in OCCI Entity collection', line)
            locations.append(uri)

        return locations


    def parse_resource(self, body, header):
        """Parse OCCI Resource instance

        :param string body[]: text to parse
        :param string headers[]: headers to parse (unused in text/plain)
        :return: categories, links, and attributes
        :rtype: [occi.Category categories[], occi.Link links[], occi.Attribute attributes[]]
        """
        categories = []
        links = []
        attributes = []

        for line in body:
            if not line.strip():
                continue

            line = line.rstrip('\r\n')
            matched = TextRenderer.reCategory.match(line)
            if matched is not None:
                s = line[matched.end():]
                categories.append(self.parse_category_body(s))
                continue
            matched = TextRenderer.reLink.match(line)
            if matched is not None:
                s = line[matched.end():]
                links.append(self.parse_link_body(s))
                continue
            matched = TextRenderer.reAttribute.match(line)
            if matched is not None:
                s = line[matched.end():]
                attributes.append(self.parse_attribute_body(s))
                continue
            else:
                raise occi.ParseError('Unexpected content of OCCI Resource instance')

        return [categories, links, attributes]



import urlparse

import occi


def check_url(body, scheme=False, host=False, path=False):
    """Check validity of URL.

    :param string body: validated URL
    :param bool scheme: scheme required
    :param bool host: hostname required
    :param bool path: local path required
    :return: URL validity, empty string considered as not valid
    :rtype: bool
    """
    url = urlparse.urlparse(body)

    if not url:
        return False

    if scheme and not url.scheme:
        return False
    if host and not url.netloc:
        return False
    if path and not url.path:
        return False
    if not url.scheme and not url.netloc and not url.path:
        return False

    return True


class Renderer:
    """ OCCI Renderer base skeleton.
    """

    def render_category(self, category):
        """Render OCCI Category

        :param occi.Category category: OCCI Category object
        :return: render result
        :rtype: [string, string[]]
        """

    def render_categories(self, categories):
        """Render OCCI Category collection

        :param occi.Category category[]: OCCI Category array
        :return: render result
        :rtype: [string, string[]]
        """


    def render_resource(self, categories, links=None, attributes=None):
        """Render OCCI Resource instance

        :param occi.Category categories[]: OCCI Category array
        :param occi.Link links[]: OCCI Link array
        :param occi.Attribute attributes[]: OCCI Attribute array
        :return: render result
        :rtype: [string, string[]]
        """
        if categories is None or not categories:
            raise occi.RenderError('Category required')


    def parse_categories(self, body, headers):
        """Parse OCCI Category Collection

        :param string body[]: text to parse
        :param string headers[]: headers to parse
        :return: Array of OCCI Categories
        :rtype: occi.Category[]
        """


    def parse_locations(self, body, headers):
        """Parse OCCI Entity collection

        :param string body[]: text to parse
        :param string headers[]: headers to parse
        :return: array of renderer-specific strings
        :rtype: string[]
        """


    def parse_resource(self, body, header):
        """Parse OCCI Resource instance

        :param string body[]: text to parse
        :param string headers[]: headers to parse
        :return: categories, links, and attributes
        :rtype: [occi.Category categories[], occi.Link links[], occi.Attribute attributes[]]
        """



# Always prefer setuptools over distutils
from setuptools import setup, find_packages
# To use a consistent encoding
from codecs import open
from os import path

import pOCCI

here = path.abspath(path.dirname(__file__))

# Get the long description from the relevant file
with open(path.join(here, 'DESCRIPTION.rst'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='pOCCI',

    version=pOCCI.__version__,

    description='OCCI Compliance Testing Tool',
    long_description=long_description,

    # The project's main homepage.
    url='https://github.com/CESNET/pOCCI',

    # Author details
    author='CESNET',
    author_email='cloud@rt4.cesnet.cz',

    license='MIT',

    classifiers=[
        'Development Status :: 3 - Alpha',

        'Intended Audience :: Developers',
        'Intended Audience :: Information Technology',
        'Topic :: Scientific/Engineering :: Interface Engine/Protocol Translator',
        'Topic :: System :: Distributed Computing',

        'License :: OSI Approved :: MIT License',

        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
    ],

    keywords='OCCI cloud compliance testing pycurl',

    packages=find_packages(exclude=['doc', 'tests']),

    install_requires=['pycurl'],

    package_data={
        'pOCCI': [
            'pOCCI.cfg',
            'pOCCI.1',
            'pOCCI-parse.1',
        ],
        'tests': ['*/*.txt'],
    },

    entry_points={
        'console_scripts': [
            'pOCCI=pOCCI.pOCCI:main',
            'pOCCI-parse=pOCCI.pOCCI_parse:main',
        ],
    },

    test_suite='tests',

    use_2to3=True,
)



import re

from render_json import *
from render_http import *
from render_text import *
from render_uri import *


mimetypes = [
    'text/plain',
    'text/occi+plain',
    'text/occi',
    'text/uri-list',
    'application/occi+json',
]


def parse_mimetype(mimetype):
    """ Parse mimetype name string.

    Example::
       text/plain; charset=iso-8859-1

    :param string mimetype: mimetype
    :result: (mimetype, charset) tuple
    :rtype: (string, string)
    """
    charset = None

    m = re.split(r';\s*', mimetype, 2)

    if len(m) > 2:
        return (None, None)

    if len(m) == 2:
        charset = m[1]

    return (m[0], charset)


def create_renderer(req_mimetype):
    """Create OCCI Renderer.

    :param string req_mimetype: requested mimetype
    :return: renderer or None
    :rtype: RenderBase
    """
    renderer = None
    mimetype, charset = parse_mimetype(req_mimetype)

    if mimetype in ['text/plain', 'text/occi+plain']:
        renderer = TextRenderer()
    elif mimetype in ['text/occi']:
        renderer = HTTPHeadersRenderer()
    elif mimetype in ['text/uri-list']:
        renderer = URIListRenderer()
    elif mimetype in ['application/occi+json']:
        renderer = JSONRenderer()

    return renderer



#!/usr/bin/python

import getopt
import re
import sys
import time
import os

from occi_libs import *
from CORE import *
import occi
import version


tests_definitions = {
    'OCCI/CORE/DISCOVERY/001': CORE_DISCOVERY001,
    'OCCI/CORE/DISCOVERY/002': CORE_DISCOVERY002,
    'OCCI/CORE/READ/001': CORE_READ001,
    'OCCI/CORE/READ/002': CORE_READ002,
    'OCCI/CORE/READ/007': CORE_READ007,
    'OCCI/CORE/CREATE/001': CORE_CREATE001,
    'OCCI/CORE/CREATE/006': CORE_CREATE006,
    'OCCI/CORE/DELETE/001': CORE_DELETE001,
    'OCCI/CORE/UPDATE/001': CORE_UPDATE001,
    'OCCI/CORE/MISC/001': CORE_MISC001,

    'OCCI/INFRA/CREATE/001': INFRA_CREATE001,
    'OCCI/INFRA/CREATE/002': INFRA_CREATE002,
    'OCCI/INFRA/CREATE/003': INFRA_CREATE003,
    'OCCI/INFRA/CREATE/004': INFRA_CREATE004,
    'OCCI/INFRA/CREATE/005': INFRA_CREATE005,
    'OCCI/INFRA/CREATE/006': INFRA_CREATE006,
    'OCCI/INFRA/CREATE/007': INFRA_CREATE007,
}


tests_skipped = set([
    'OCCI/CORE/CREATE/001',
    'OCCI/CORE/CREATE/006',
    'OCCI/INFRA/CREATE/005',
])


def usage(name=__file__):
    print "%s [OPTIONS]\n\
\n\
OPTIONS:\n\
  -h, --help ................ usage message\n\
  -a, --auth-type AUTH ...... authentization type\n\
  -c, --cert FILE ........... SSL certificate file\n\
  -e, --endpoint, --url URL . OCCI server endpoint\n\
  -f, --format FORMAT ....... output format (plain, json)\n\
  -k, --key FILE ............ SSL key file\n\
  -l, --list ................ list all test\n\
  -m, --mime-type MIME-TYPE . render format\n\
  -p, --password PWD ........ password for basic auth-type\n\
  -P, --passphrase PASS ..... SSL key passphrase\n\
  -s, --keystone TENANT ..... tenant used for keystone auth\n\
  -S, --ignore-ssl .......... ignore SSL errors\n\
  -t, --tests <TEST1,...> ... list of tests\n\
  -u, --user USER ........... user name for basic auth-type\n\
  -v, --verbose ............. verbose mode\n\
  -V, --version ............. print version information\n\
\n\
MIME-TYPE: text/plain, text/occi\n\
AUTH: basic\n\
" % os.path.basename(name)


def main(argv=sys.argv[1:]):
    results = []
    tests = []

    if not occi_config:
        print >> sys.stderr, 'No configuration found'
        sys.exit(2)

    try:
        opts, args = getopt.getopt(argv, "ha:c:e:f:k:lm:p:Ps::St:u:vV", ["help", "auth-type=", "cert=", "endpoint=", "format=", "ignore-ssl", "keystone=", "key=", "list", "mime-type=", "passphrase=", "password=", "tests=", "url=", "user=", "verbose", "version"])
    except getopt.GetoptError:
        usage()
        sys.exit(2)
    for opt, arg in opts:
        if opt in ["-h", "--help"]:
            usage()
            sys.exit()
        elif opt in ["-a", "--auth-type"]:
            occi_config["authtype"] = arg
        elif opt in ["-c", "--cert"]:
            occi_config["cert"] = arg
        elif opt in ["-e", "--endpoint", "--url"]:
            occi_config["url"] = arg
        elif opt in ["-f", "--format"]:
            occi_config['outputformat'] = arg
        elif opt in ["-k", "--key"]:
            occi_config["key"] = arg
        elif opt in ["-l", "--list"]:
            print '\n'.join(sorted(tests_definitions.keys()))
            sys.exit()
        elif opt in ["-m", "--mime-type"]:
            occi_config['mimetype'] = arg
        elif opt in ["-P", "--passphrase"]:
            occi_config["passphrase"] = arg
        elif opt in ["-p", "--password"]:
            occi_config["passwd"] = arg
        elif opt in ["-s", "--keystone"]:
            occi_config["keystone"] = True
        elif opt in ["-S", "--ignore-ssl"]:
            occi_config["ignoressl"] = True
        elif opt in ["-t", "--tests"]:
            tests = arg.split(",")
        elif opt in ["-u", "--user"]:
            occi_config["user"] = arg
        elif opt in ["-v", "--verbose"]:
            occi_config['curlverbose'] = True
        elif opt in ["-V", "--version"]:
            print version.__version__
            sys.exit()

    if not tests:
        tests = sorted(list(set(tests_definitions.keys()) - tests_skipped))

    if 'url' not in occi_config or not occi_config['url']:
        print 'OCCI URL required'
        sys.exit(2)

    if occi_config['authtype'] == 'basic':
        if 'user' not in occi_config or not occi_config['user']:
            print 'User and password is required for "basic" authentization type.'
            sys.exit(2)
    elif occi_config['authtype'] == 'x509':
        if 'X509_USER_PROXY' in os.environ:
            if 'cert' not in occi_config or not occi_config['cert']:
                occi_config['cert'] = os.environ['X509_USER_PROXY']
            if 'cachain' not in occi_config or not occi_config['cachain']:
                occi_config['cachain'] = os.environ['X509_USER_PROXY']
        if 'cert' not in occi_config or not occi_config['cert']:
            print 'SSL certificate and key is required for "x509" authentization type.'
            sys.exit(2)
        if 'capath' not in occi_config or not occi_config['capath']:
            if 'X509_CERT_DIR' in os.environ:
                occi_config['capath'] = os.environ['X509_CERT_DIR']
            elif os.path.isdir('/etc/grid-security/certificates'):
                occi_config['capath'] = '/etc/grid-security/certificates'


    occi_init()
    testsuite_init()

    for test in tests:
        if not re.match(r'OCCI/', test):
            test = 'OCCI/' + test
        if test in tests_definitions.keys():
            start_time = time.time()
            try:
                result, err_msg = tests_definitions[test].test()
            except occi.Error as oe:
                if occi_config['curlverbose']:
                    raise oe
                else:
                    print '%s error: %s' % (test, str(oe))
                    sys.exit(2)
            objective = tests_definitions[test].objective
            running_time = time.time() - start_time
            results.append(occi_test(test, objective, result, err_msg, running_time))

    results = occi_format(results)
    occi_print(results, occi_config['outputformat'])

    if results['failed'] == 0:
        sys.exit(0)
    else:
        sys.exit(1)


if __name__ == "__main__":
    main(sys.argv[1:])
